//############################################################################
//##                                                                        ##
//##  FPMATH.H: Vector and matrix math routines                             ##
//##                                                                        ##
//############################################################################
//##                                                                        ##
//## Released to the public domain 12/15/99 by John Miles (jmiles@pop.net)  ##
//## No restrictions are placed on use or distribution of any of this code. ##
//##                                                                        ##
//############################################################################

#ifndef FPMATH_H
#define FPMATH_H

//
// Disable warning to allow literal float constants without 'F' appelation
//

#pragma warning(disable:4244)

#include "typedefs.h"

#ifdef __cplusplus
extern "C" {
#endif

//
// Definitions
//

#ifndef PI
  #define PI 3.1415926535
#endif

#ifndef MUL_DEG_TO_RAD
  #define MUL_DEG_TO_RAD ((2.0 * PI) / 360.0)
#endif

//
// Structures
//

typedef struct
{
   SINGLE x;
   SINGLE y;
   SINGLE z;
}
VECTOR;

//
// Rotation axis definitions
//

#define X_AXIS 0
#define Y_AXIS 1
#define Z_AXIS 2

#define PITCH  X_AXIS
#define YAW    Y_AXIS
#define ROLL   Z_AXIS

//
// 4x3 matrix transform
//

typedef struct
{
   SINGLE element[3][4]; // Matrix in [row][col] order; 4th row is { 0 0 0 1 }
}
TRANSFORM;

//
// Identity matrix (X*1.0, Y*1.0, Z*1.0, all other elements 0)
//

extern const TRANSFORM identity;

//
// Function prototypes
// 

extern void       VECTOR_rotate                  (VECTOR *src, VECTOR *dest, TRANSFORM *r);
extern void       VECTOR_rotate_translate        (VECTOR *src, VECTOR *dest, TRANSFORM *r);
extern void       VECTOR_inverse_rotate          (VECTOR *src, VECTOR *dest, TRANSFORM *r);
extern void       VECTOR_inverse_rotate_translate(VECTOR *src, VECTOR *dest, TRANSFORM *r);
extern S32        VECTOR_compare                 (VECTOR *A,   VECTOR *B,    SINGLE tolerance);
extern SINGLE     VECTOR_magnitude               (VECTOR *V);
extern SINGLE     VECTOR_square_magnitude        (VECTOR *V);
extern SINGLE     VECTOR_displacement            (VECTOR *V, VECTOR *W);
extern SINGLE     VECTOR_square_displacement     (VECTOR *V, VECTOR *W);
extern void       TRANSFORM_multiply             (TRANSFORM *M1, TRANSFORM *M2, TRANSFORM *out);
extern void       TRANSFORM_compose_rotation     (TRANSFORM *X,  U32 axis,      SINGLE angle);
extern TRANSFORM *TRANSFORM_transpose            (TRANSFORM *X);
extern TRANSFORM *TRANSFORM_invert               (TRANSFORM *X);
extern void       TRANSFORM_get_position         (TRANSFORM *X,  VECTOR    *V);
void              TRANSFORM_set_position         (TRANSFORM *X, SINGLE x, SINGLE y, SINGLE z);
void              TRANSFORM_move_position        (TRANSFORM *X, SINGLE dx, SINGLE dy, SINGLE dz);
extern void       TRANSFORM_set_orientation      (TRANSFORM *X, SINGLE pitch, SINGLE roll, SINGLE yaw);
extern S32        TRANSFORM_load_from_file       (TRANSFORM *X, C8 *filename);

#ifdef __cplusplus
}
#endif
   
#endif


