//############################################################################
//##                                                                        ##
//##  openFLAME                                                             ##
//##                                                                        ##
//##  OBJECT.H: Base class for all tangible (mesh-derived) 3D objects       ##
//##            handled by openFLAME                                        ##
//##                                                                        ##
//############################################################################
//##                                                                        ##
//## Released to the public domain 12/15/99 by John Miles (jmiles@pop.net)  ##
//## No restrictions are placed on use or distribution of any of this code. ##
//##                                                                        ##
//############################################################################

#ifndef OBJECT_H
#define OBJECT_H

#ifdef __cplusplus
extern "C" {
#endif

//
// Object representation
//

typedef struct
{
   TRANSFORM   to_world;            // Object space -> world space
   MESH       *mesh;                // Mesh template for construction of object

   //
   // Frame-coherent visibility information
   //

   S32         current_leaf;        // Last rendered POV was in this BSP leaf

   U32        *face_marked_visible; // Last cycle in which this face was 
                                    // marked visible

   U32         face_mark_cycle;     // Last cycle during which this object's
                                    // faces were marked for visibility

   S32         visibility_valid;    // FALSE if no visibility information available
}
OBJECT;

//
// Prototypes
//

extern OBJECT *   OBJECT_create (MESH     *mesh);
extern void       OBJECT_destroy(OBJECT   *object);

extern void       OBJECT_render (OBJECT    *object,
                                 TRANSFORM *view_to_world);

extern S32        BSP_classify_point(OBJECT *obj,
                                     SINGLE  x,
                                     SINGLE  y,
                                     SINGLE  z);
#ifdef __cplusplus
}
#endif
   
#endif
