//############################################################################
//##                                                                        ##
//##  Q.CPP                                                                 ##
//##                                                                        ##
//##  Test bed                                                              ##
//##                                                                        ##
//##  V1.00 of 19-Aug-99: Initial                                           ##
//##                                                                        ##
//##  Author: John Miles                                                    ##
//##                                                                        ##
//##  C++ source compatible with MSVC 12                                    ##
//##                                                                        ##
//############################################################################
//##                                                                        ##
//## Released to the public domain 12/15/99 by John Miles (jmiles@pop.net)  ##
//## No restrictions are placed on use or distribution of any of this code. ##
//##                                                                        ##
//############################################################################

#include <string.h>
#include <stdio.h>
#include <dos.h>
#include <stdlib.h>
#include <ctype.h>
#include <math.h>

#include "typedefs.h"
#include "userio.h"
#include "system.h"

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>

#include <richedit.h>
#include <commctrl.h>
#include <commdlg.h>

#include "mesh.h"
#include "object.h"
#include "fpmath.h"

//
// Globals
//

const C8 VERSION[] = "1.00";

USERIO UIO;

HGLFONT   fnt = -1;
HWINDOW   primary = -1;
GLVTABLE *GL;

USERWINMODE requested_mode;
USERWINMODE current_mode;
S32         mode_change_request;

#define GL_PROVIDER "opengl32.dll"

#define RES_X 640
#define RES_Y 480

SINGLE   fx,fy;
SINGLE   yaw,pitch;
TRANSFORM view_to_world;
S32      left_down;
S32      right_down;
SINGLE   linear_velocity;
SINGLE   fps = 30.0F;

OBJECT      *object;
MESH        *mesh;

extern S32 wire_mode;

//****************************************************************************
//
// Window message receiver procedure for application
//
//****************************************************************************

S32 done;

long FAR PASCAL WindowProc(HWND   hWnd,   UINT   message,   //)
                           WPARAM wParam, LPARAM lParam)
{
   WINAREA area;
   POINT   point;
   S32     cx,cy,dx,dy;

   switch (message)
      {
      case WM_MOUSEMOVE:

         if (primary == -1)
            {
            break;
            }

         //
         // Get movement delta from center
         // 

         UIO.client_area(primary,
                        &area);

         cx = area.x + (area.w / 2);
         cy = area.y + (area.h / 2);

         point.x = LOWORD(lParam);
         point.y = HIWORD(lParam);

         ClientToScreen(hWnd, &point);

         dx = (cx - point.x); 
         dy = (cy - point.y); 

         if (!(dx || dy))
            {
            break;
            }

         //
         // Recenter cursor in client area
         //

         SetCursorPos(cx,cy);

         //
         // Implement MLOOK functionality
         //

         yaw += (dx / 10);

         while (yaw > 360.0)
            {
            yaw -= 360.0;
            }
         while (yaw < 0.0)
            {
            yaw += 360.0;
            }

         pitch += (dy / 5);

         while (pitch > 360.0)
            {
            pitch -= 360.0;
            }
         while (pitch < 0.0)
            {
            pitch += 360.0;
            }

         break;

      case WM_RBUTTONDOWN:

         right_down = 1;
         break;

      case WM_RBUTTONUP:

         right_down = 0;
         break;

      case WM_LBUTTONDOWN:

         left_down = 1;
         break;

      case WM_LBUTTONUP:

         left_down = 0;
         break;

      case WM_SETFOCUS:
         debug_printf("Setfocus %X %X\n",wParam,lParam);
         break;

      case WM_KILLFOCUS:
         debug_printf("Killfocus %X %X\n",wParam,lParam);
         break;

      case WM_CHAR:

         switch (wParam)
            {
            //
            // ESC terminates test
            // 

            case VK_ESCAPE:
               done = 1;
               break;

            case 'w':
            case 'W':

               wire_mode = !wire_mode;
               break;

            //
            // Any other key advances demo state
            //

            default:
               break;
            }
         break;

      case WM_SYSKEYUP:

         if (wParam == VK_RETURN)
            {
            debug_printf("INTERP: Request fullscreen/window toggle\n");
            requested_mode = (current_mode == UWM_WINDOW) ? UWM_FULLSCREEN : UWM_WINDOW;
            mode_change_request = TRUE;
            }
         break;
      }

    return DefWindowProc(hWnd, message, wParam, lParam);
}

//****************************************************************************/
//
// Set fullscreen or windowed mode
//
//****************************************************************************/

S32 set_mode(USERWINMODE mode)
{
   if (fnt != -1)
      {
      UIO.destroy_font(fnt);
      }

   if (primary != -1)
      {
      UIO.close_window(primary);
      }

   //
   // Create primary application window and load GL provider
   // 

   primary = UIO.open_window (RES_X,
                              RES_Y,
                              16,
                              mode,
                              GL_PROVIDER,
                             &GL,
                              UCW_PRIMARY);

   if (primary == -1)
      {
      return 0;
      }

   UIO.register_system_window_proc(primary, 
                                   WindowProc);

   //
   // Initialize default GL state
   //

   GL->glClearColor(0.1f, 0.2f, 0.4f, 1.0f);
   GL->glShadeModel(GL_SMOOTH);
   GL->glEnable(GL_TEXTURE_2D);
   GL->glEnable(GL_DEPTH_TEST);

   GL->glEnable(GL_BLEND);
   GL->glBlendFunc(GL_SRC_ALPHA, GL_ONE_MINUS_SRC_ALPHA);

   GL->glPixelStorei(GL_UNPACK_ALIGNMENT, 1);

   GL->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_S,     GL_REPEAT);
   GL->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_WRAP_T,     GL_REPEAT);
   GL->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MAG_FILTER, GL_NEAREST);
   GL->glTexParameteri(GL_TEXTURE_2D, GL_TEXTURE_MIN_FILTER, GL_NEAREST);
   GL->glHint(GL_PERSPECTIVE_CORRECTION_HINT, GL_NICEST);

   //
   // Set GL viewing parameters
   //

   GL->glMatrixMode(GL_PROJECTION);
   GL->glLoadIdentity();

   GL->glViewport(0, 0, RES_X, RES_Y);
   GL->glScissor (0, 0, RES_X, RES_Y);

   GL->glOrtho(0.0, RES_X, RES_Y, 0.0, -99999.0F, 99999.0F);

   GL->glMatrixMode(GL_MODELVIEW);
   GL->glLoadIdentity();

   //
   // Center cursor in client area
   //

   WINAREA area;

   UIO.client_area(primary, &area);

   SetCursorPos(area.x + (area.w / 2),
                area.y + (area.h / 2));

   //
   // Create font
   //

   LOGFONT lf;

   GetObject(GetStockObject(OEM_FIXED_FONT), sizeof(lf), &lf);

   fnt = UIO.create_font(GL,     
                         CreateFontIndirect(&lf),
                         RGB(255,255,0),
                         FNT_RENDER_AS_FIXED);

   //
   // Arrange for textures to be rebound
   // 

   MESH_invalidate_textures(mesh);

   return 1;
}

//****************************************************************************/
//
// Exit handlers must be present in every USERIO application
//
// WinExit() will be called from USERIO::serve_window() if the user hits 
// ALT-F4 or closes the window manually      
//
// AppExit() will be called if the application terminates its own 
// execution
//
//****************************************************************************/

void WINAPI WinExit(void)
{
   debug_printf("WinExit called\n");
   exit(0);
}

void AppExit(void)
{
   debug_printf("AppExit called\n");
   UIO.shutdown();

   debug_printf("Final exit OK\n");
}

//############################################################################
//#                                                                          #
//# main()                                                                   #
//#                                                                          #
//############################################################################

int WINAPI WinMain (HINSTANCE hInstance,  //)
                    HINSTANCE hPrevInstance, 
                    LPSTR lpCmdLine, 
                    int nCmdShow)
{
   //
   // Initialize system abstraction layer -- must succeed in order to continue
   //

   if (!UIO.startup(hInstance,
                   "q",
                    TRUE,
                    WinExit))
      {
      return 0;
      }

   //
   // Register exit handler 
   //

   atexit(AppExit);

   //
   // Start openFLAME
   //

   mesh = (MESH *) MESH_load_from_file(lpCmdLine);

   if (mesh == NULL)
      {
      exit(1);
      }

   //
   // Instantiate object and camera
   //

   object   = OBJECT_create(mesh);

   fx = 90.0;
   fy = 90.0;

   left_down = 0;
   right_down = 0;

   yaw   = 0.0;
   pitch = 0.0;

   TRANSFORM_load_from_file(&view_to_world, "start.xfm");

   //
   // Initialize display mode and frame counter
   //

   LARGE_INTEGER freq,last;

   QueryPerformanceFrequency(&freq);
   QueryPerformanceCounter  (&last);

   U32 CPS = freq.u.LowPart;

   UIO.hide_system_mouse();

   current_mode = requested_mode = UWM_WINDOW;
   mode_change_request = TRUE;

   primary = -1;

   while (!done)
      {
      if (mode_change_request)
         {
         if (!set_mode(requested_mode))
            {
            return 0;
            }

         current_mode = requested_mode;
         mode_change_request = FALSE;
         }

      UIO.serve_window(primary);

      GL->glDisable(GL_CULL_FACE);
      GL->glDisable(GL_CLIP_PLANE0);
      GL->glDisable(GL_DEPTH_TEST);
      GL->glPolygonMode(GL_FRONT_AND_BACK, GL_FILL);

      //
      // Set up to render scene
      //

      linear_velocity = 0.0;

      if (left_down)
         {
         linear_velocity = 450.0;
         }
      else if (right_down)
         {
         linear_velocity = -450.0;      // units per second
         }

      linear_velocity /= fps;           // div fps = units per frame

      TRANSFORM_move_position(&view_to_world, 0.0, 0.0, -linear_velocity);
      TRANSFORM_set_orientation(&view_to_world, pitch, 0.0, yaw);

      GL->glClear(GL_COLOR_BUFFER_BIT | GL_DEPTH_BUFFER_BIT);

      GL->glMatrixMode(GL_PROJECTION);
      GL->glLoadIdentity();

      GL->glViewport(0, 0, RES_X, RES_Y);
      GL->glScissor (0, 0, RES_X, RES_Y);

      GL->glFrustum(-4.000000F, 
                     4.000000F, 
                    -3.000000F,
                     3.000000F,
                     4.000000F, 
                     4096.000000F);

      OBJECT_render(object,                         // Object to render
                   &view_to_world);                 // Camera

      //
      // Show frame rate
      //

      GL->glMatrixMode(GL_PROJECTION);
      GL->glLoadIdentity();

      GL->glViewport(0, 0, RES_X, RES_Y);
      GL->glScissor (0, 0, RES_X, RES_Y);

      GL->glOrtho(0.0, RES_X, RES_Y, 0.0, -99999.0F, 99999.0F);

      GL->glMatrixMode(GL_MODELVIEW);
      GL->glLoadIdentity();

      LARGE_INTEGER now;
      C8 strbuf[128];

      QueryPerformanceCounter(&now);
      U32 elapsed = now.u.LowPart - last.u.LowPart;
      last = now;

      fps = ((SINGLE) CPS) / ((SINGLE) elapsed);

      sprintf(strbuf,"FPS: %3.1f   ",fps);

      UIO.draw_string(fnt,strbuf,290,30);

      //
      // Display rendered scene
      //

      SwapBuffers(GL->hDC);
      }

   return 0;
}
