#ifndef QUAKE_HEADER
#define QUAKE_HEADER

//******************************************************
//** QUAKE Specific Structures From Quake BSP Headers **
//******************************************************

#define LUMP_ENTITIES      0
#define LUMP_PLANES        1
#define LUMP_TEXTURES      2
#define LUMP_VERTEXES      3
#define LUMP_VISIBILITY    4
#define LUMP_NODES         5
#define LUMP_TEXINFO       6
#define LUMP_FACES         7
#define LUMP_LIGHTING      8
#define LUMP_CLIPNODES     9
#define LUMP_LEAFS        10
#define LUMP_MARKSURFACES 11
#define LUMP_EDGES        12
#define LUMP_SURFEDGES    13
#define LUMP_MODELS       14
#define HEADER_LUMPS      15

#define BSPVERSION        29

#define MAX_MAP_HULLS      4
#define MAX_MAP_LEAFS  32767

#define MIPLEVELS          4

#define PLANE_X            0
#define PLANE_Y            1
#define PLANE_Z            2
#define PLANE_ANYX         3
#define PLANE_ANYY         4
#define PLANE_ANYZ         5

#define CONTENTS_EMPTY    -1
#define CONTENTS_SOLID    -2
#define CONTENTS_WATER    -3
#define CONTENTS_SLIME    -4
#define CONTENTS_LAVA     -5
#define CONTENTS_SKY      -6

#define MAXLIGHTMAPS       4

#define AMBIENT_WATER      0
#define AMBIENT_SKY        1
#define AMBIENT_SLIME      2
#define AMBIENT_LAVA       3
#define NUM_AMBIENTS       4

typedef struct
 {
 S32 fileofs,filelen;
 }
lump_t;

typedef struct
 {
 S32    version; 
 lump_t lumps[HEADER_LUMPS];
 }
dheader_t;

typedef struct
 {
 SINGLE mins[3],maxs[3];
 SINGLE origin[3];
 S32    headnode[MAX_MAP_HULLS];
 S32    visleafs;
 S32    firstface,numfaces;
 }
dmodel_t;

typedef struct
 {
 S32 nummiptex;
 S32 dataofs[4];
 }
dmiptexlump_t;

typedef struct miptex_s
 {
 C8  name[16];
 U32 width,height;
 U32 offsets[MIPLEVELS];
 }
miptex_t;

typedef struct
 {
 SINGLE point[3];
 }
dvertex_t;

typedef struct
 {
 SINGLE normal[3];
 SINGLE dist;
 S32    type;
 }    
dplane_t;

typedef struct
 {
 S32 planenum;
 S16 children[2];
 S16 mins[3];
 S16 maxs[3];
 U16 firstface;
 U16 numfaces;
 }
dnode_t;

typedef struct
 {
 S32 contents;
 S32 visofs;
 S16 mins[3];
 S16 maxs[3];
 U16 firstmarksurface;
 U16 nummarksurfaces;
 U8  ambient_level[NUM_AMBIENTS];
 }
dleaf_t;

typedef struct
 {
 S32 planenum;
 S16 children[2];
 }
dclipnode_t;

typedef struct texinfo_s
 {
 SINGLE vecs[2][4];
 S32    miptex;
 S32    flags;
 }
texinfo_t;

typedef struct
 {
 U16 v[2];
 }
dedge_t;

typedef struct
 {
 S16 planenum;
 S16 side;
 S32 firstedge;
 S16 numedges; 
 S16 texinfo;
 U8  styles[MAXLIGHTMAPS];
 S32 lightofs;
 }
dface_t;

//*******************************
//** The Super Combo Structure **
//*******************************

typedef struct
 {
 dheader_t    header;
 S32          numEntityData;
 S32          numPlanes;
 S32          numTextureData;
 S32          numVerts;
 S32          numVisData;
 S32          numNodes;
 S32          numTexInfo;
 S32          numFaces;
 S32          numLightingData;
 S32          numClipNodes;
 S32          numLeafs;
 S32          numMarkSurfaces;
 S32          numEdges;
 S32          numSurfEdges;
 S32          numModels;
 U8          *entityData;
 dplane_t    *plane;
 U8          *textureData;
 dvertex_t   *vert;
 U8          *visData;
 dnode_t     *node;
 texinfo_t   *texInfo;
 dface_t     *face;
 U8          *lightingData;
 dclipnode_t *clipNode;
 dleaf_t     *leaf;
 U16         *markSurface;
 dedge_t     *edge;
 S32         *surfEdge;
 dmodel_t    *model;
 }
QUAKE;

#endif

