//****************************************************************************
//*                                                                          *
//* SYSTEM.CPP: General-purpose system routines (non-member)                 *
//*                                                                          *
//* 32-bit protected-mode source compatible with MSVC 10.2                   *
//*                                                                          *
//* Version 1.00 of 29-Jun-99: Initial                                       *
//*                                                                          *
//* Author: John Miles                                                       *
//*                                                                          *
//############################################################################
//##                                                                        ##
//## Released to the public domain 12/15/99 by John Miles (jmiles@pop.net)  ##
//## No restrictions are placed on use or distribution of any of this code. ##
//##                                                                        ##
//############################################################################

#include <stdio.h>
#include <string.h>
#include <malloc.h>
#include <stdlib.h>
#include <stdarg.h>
#include <time.h>

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <mmsystem.h>

#include "system.h"

S32 log_to_file             = 0;
C8  debug_log_filename[512] = "";
U32 debug_start_time        = 0;
FILE *debugfile = NULL;
S32 debugfile_initialized = 0;

//****************************************************************************
//*                                                                          *
//*  Monochrome debug printf() function                                      *
//*                                                                          *
//*  Text written to logfile if option enabled                               *
//*                                                                          *
//****************************************************************************

void __cdecl debug_printf(C8 *fmt, ...)
{
   static C8 work_string[4096];
   FILE     *log;

   //
   // Force floating-point inclusion so debug_printf() can print floats
   //

   volatile float fval;
   fval = 1.0F;

   //
   // Check for debugging logfile request (environment variable in DOS 
   // version, or WIN.INI entry in Windows version)
   //
   // If file logging enabled, append session date/timestamp to debug logfile
   //

   if (!debugfile_initialized)
      {
      debugfile_initialized = 1;

      GetProfileString("Eden", "EdenDebug", "None", debug_log_filename, sizeof(debug_log_filename));

      if (!strnicmp(debug_log_filename, "None", 4))
         {
         log_to_file = FALSE;
         }
      else
         {
         //
         // Create script file and set "debug" flag
         //
         // A leading '+' allows debug output to be appended to existing file
         //

         if (debug_log_filename[0] == '+')
            {
            strcpy(debug_log_filename, &debug_log_filename[1]);
            debugfile = fopen(debug_log_filename,"a+t");
            }
         else
            {
            debugfile = fopen(debug_log_filename,"w+t");
            }

         if (debugfile == NULL)
            {
            debug_printf("Eden: Unable to open debug logfile \"%s\"\n",
               debug_log_filename);
            }
         else
            {
            //
            // Write header to log file
            //

            S32        elapstime;
            struct tm *adjtime;
            C8         loctime[32];

            time(&elapstime);
            adjtime = localtime(&elapstime);
            strcpy(loctime,asctime(adjtime));
            loctime[24] = 0;

            debug_start_time = timeGetTime();

            fprintf(debugfile,
                  "-------------------------------------------------------------------------------\n"
                  "Eden debug script\n"
                  "Start time: %s\n"
                  "-------------------------------------------------------------------------------\n\n"
                  ,(C8 FAR *) loctime);

            log_to_file = TRUE;
            fclose(debugfile);
            }
         }
      }

   if ((fmt == NULL) || (strlen(fmt) > sizeof(work_string)))
      {
      strcpy(work_string, "(String missing or too large)");
      }
   else
      {
      va_list ap;

      va_start(ap, 
               fmt);

      vsprintf(work_string, 
               fmt, 
               ap);

      va_end  (ap);
      }

   if ((log_to_file) && (strlen(work_string)))
      {
      static S32 previous_newline = 1;

      log = fopen(debug_log_filename,"a+t");
      
      if (log != NULL)
         {
         if (previous_newline)
            {
            U32 t = timeGetTime() - debug_start_time;

            fprintf(log,"[%.02d:%.02d:%.02d.%.03d] %s",
               (t / 3600000L),
               ((t / 60000L) % 60L),
               ((t / 1000L)  % 60L),
               (t % 1000L),
               work_string);
            }
         else
            {
            fprintf(log,"%s",work_string);
            }

         fclose(log);

         previous_newline = (work_string[strlen(work_string)-1] == '\n');
         }
      }

   OutputDebugString(work_string);
}

//****************************************************************************
//*                                                                          *
//*  Message box display function                                            *
//*                                                                          *
//****************************************************************************

void __cdecl alert_box(C8 *caption, C8 *fmt, ...)
{
   static C8 work_string[4096];

   ShowCursor(1);

   if ((fmt == NULL) || (strlen(fmt) > sizeof(work_string)))
      {
      strcpy(work_string, "(String missing or too large)");
      }
   else
      {
      va_list ap;

      va_start(ap, 
               fmt);

      vsprintf(work_string, 
               fmt, 
               ap);

      va_end  (ap);
      }

   debug_printf("%s\n",work_string);

//   //
//   // If DirectDraw active, display GDI surface for dialog box
//   // 
//
//   if (DDraw_active)
//      {
//      DDraw.lpDD->FlipToGDISurface();
//      }

   //
   // If in 8-bpp mode, save current palette and switch to standard palette
   // 

   if (caption == NULL)
      {
      MessageBox(NULL,        // was hWnd
                 work_string,
                 "SAL Error", 
                 MB_OK);
      }
   else
      {
      MessageBox(NULL, 
                 work_string,
                 caption,
                 MB_OK);
      }

   ShowCursor(0);
}

//************************************************************/
//
// Return base-2 logarithm of unsigned value, or -1 if value 
// is not an integral power of 2
//
//************************************************************/

S32 log2(U32 value)
{
   S32 n;
   U32 pwr;

   for (pwr=1,n=0; n < 32; n++)
      {
      if (value == pwr)
         {
         return n;
         }

      pwr <<= 1;
      }

   return -1;
}

//************************************************************/
//
// memset() clone for U32 values
//
//************************************************************/

U32  *fill_U32  (void *dest, //)
                 U32   ch,
                 U32   cnt)
{
   U32 *d = (U32 *) dest;

   while (cnt--) *d++ = ch;

   return (U32 *) dest;
}


