//############################################################################
//##                                                                        ##
//##  W32GL.CPP: USERIO implementation for Win32 OpenGL                     ##
//##                                                                        ##
//##  Version 1.00 of 30-Jul-99: Initial, derived from W32SAL 2.00          ##
//##                                                                        ##
//##  Project: Eden runtime interpreter                                     ##
//##   Author: John Miles                                                   ##
//##                                                                        ##
//##  C++ source compatible with MSVC 12                                    ##
//##                                                                        ##
//############################################################################
//##                                                                        ##
//## Released to the public domain 12/15/99 by John Miles (jmiles@pop.net)  ##
//## No restrictions are placed on use or distribution of any of this code. ##
//##                                                                        ##
//############################################################################

#define WIN32_LEAN_AND_MEAN
#include <windows.h>
#include <windowsx.h>
#include <winuser.h>
#include <mmsystem.h>
#include <gl/gl.h>

#include <stdio.h>
#include <conio.h>
#include <dos.h>
#include <malloc.h>
#include <direct.h>
#include <io.h>
#include <time.h>
#include <assert.h>

#include "system.h"
#include "userio.h"
#include "stdtpl.h"

struct WINDOW
{
   U32     hash_key;          // For internal use only
   WINDOW *hash_next;   
   WINDOW *hash_prev;  
   WINDOW *next;          
   WINDOW *prev;       
                             
   S32     index;             // Index of this entry in linear array

   static U32 hash  (void *V) { return (U32(V) & 0xff0) >> 4; } // Derive hash key from object
   BOOL32 compare   (void *V) { return (V == hWnd); }           // Compare object with object stored in T
   void   initialize(void *V) { hWnd = (HWND) V;    }           // Store indicated object in T
   void   shutdown  (void)    {}                                // Called when unlinked from pool
   void   display   (void)    {}                                // Optionally display contents of T

   //
   // User vars
   //

   HWND        hWnd;
   S32         setting_fullscreen_mode;

   WINAREA     area;                     // Location/size of window client area (windowed mode)
   RECT        original_window_rect;     // Default size of window
   WINAREA     original_area;            // Default size of client area
               
   HGLRC       GL_RC;                    // OpenGL rendering context for this thread
   GLVTABLE    GL;                       // OpenGL provider information
   S32         GL_fullscreen;            // OpenGL has been initialized in fullscreen mode
   C8          GL_name[MAX_PATH];        // Name of provider library
   
   S32         display_width;            // Display size and BPP
   S32         display_height;
   S32         display_bpp;
   USERWINMODE window_mode;              // Windowed/fullscreen mode
   U32         flags;                    // UCW_ flags

   WNDPROC     callback;                 // User window message callback

   S32 quit_message_received;
   S32 manually_destroying;
};

//
// Globals
// 

S32 app_active;

HINSTANCE  hAppInstance;              // Application instance handle
HINSTANCE  hDLLInstance;              // DLL instance handle
char       szAppName[512];            // Application name

USEREXITCB exit_callback;             // Address of mandatory exit callback

HashPool <WINDOW, 256, 256> windows;  // List of managed windows, hashed for fast lookup

//
// Names of OpenGL functions
//

const static C8 *GLN[] = 
{
   "glAccum",
   "glAlphaFunc",
   "glAreTexturesResident",
   "glArrayElement",
   "glBegin",
   "glBindTexture",
   "glBitmap",
   "glBlendFunc",
   "glCallList",
   "glCallLists",
   "glClear",
   "glClearAccum",
   "glClearColor",
   "glClearDepth",
   "glClearIndex",
   "glClearStencil",
   "glClipPlane",
   "glColor3b",
   "glColor3bv",
   "glColor3d",
   "glColor3dv",
   "glColor3f",
   "glColor3fv",
   "glColor3i",
   "glColor3iv",
   "glColor3s",
   "glColor3sv",
   "glColor3ub",
   "glColor3ubv",
   "glColor3ui",
   "glColor3uiv",
   "glColor3us",
   "glColor3usv",
   "glColor4b",
   "glColor4bv",
   "glColor4d",
   "glColor4dv",
   "glColor4f",
   "glColor4fv",
   "glColor4i",
   "glColor4iv",
   "glColor4s",
   "glColor4sv",
   "glColor4ub",
   "glColor4ubv",
   "glColor4ui",
   "glColor4uiv",
   "glColor4us",
   "glColor4usv",
   "glColorMask",
   "glColorMaterial",
   "glColorPointer",
   "glCopyPixels",
   "glCopyTexImage1D",
   "glCopyTexImage2D",
   "glCopyTexSubImage1D",
   "glCopyTexSubImage2D",
   "glCullFace",
   "glDeleteLists",
   "glDeleteTextures",
   "glDepthFunc",
   "glDepthMask",
   "glDepthRange",
   "glDisable",
   "glDisableClientState",
   "glDrawArrays",
   "glDrawBuffer",
   "glDrawElements",
   "glDrawPixels",
   "glEdgeFlag",
   "glEdgeFlagPointer",
   "glEdgeFlagv",
   "glEnable",
   "glEnableClientState",
   "glEnd",
   "glEndList",
   "glEvalCoord1d",
   "glEvalCoord1dv",
   "glEvalCoord1f",
   "glEvalCoord1fv",
   "glEvalCoord2d",
   "glEvalCoord2dv",
   "glEvalCoord2f",
   "glEvalCoord2fv",
   "glEvalMesh1",
   "glEvalMesh2",
   "glEvalPoint1",
   "glEvalPoint2",
   "glFeedbackBuffer",
   "glFinish",
   "glFlush",
   "glFogf",
   "glFogfv",
   "glFogi",
   "glFogiv",
   "glFrontFace",
   "glFrustum",
   "glGenLists",
   "glGenTextures",
   "glGetBooleanv",
   "glGetClipPlane",
   "glGetDoublev",
   "glGetError",
   "glGetFloatv",
   "glGetIntegerv",
   "glGetLightfv",
   "glGetLightiv",
   "glGetMapdv",
   "glGetMapfv",
   "glGetMapiv",
   "glGetMaterialfv",
   "glGetMaterialiv",
   "glGetPixelMapfv",
   "glGetPixelMapuiv",
   "glGetPixelMapusv",
   "glGetPointerv",
   "glGetPolygonStipple",
   "glGetString",
   "glGetTexEnvfv",
   "glGetTexEnviv",
   "glGetTexGendv",
   "glGetTexGenfv",
   "glGetTexGeniv",
   "glGetTexImage",
   "glGetTexLevelParameterfv",
   "glGetTexLevelParameteriv",
   "glGetTexParameterfv",
   "glGetTexParameteriv",
   "glHint",
   "glIndexMask",
   "glIndexPointer",
   "glIndexd",
   "glIndexdv",
   "glIndexf",
   "glIndexfv",
   "glIndexi",
   "glIndexiv",
   "glIndexs",
   "glIndexsv",
   "glIndexub",
   "glIndexubv",
   "glInitNames",
   "glInterleavedArrays",
   "glIsEnabled",
   "glIsList",
   "glIsTexture",
   "glLightModelf",
   "glLightModelfv",
   "glLightModeli",
   "glLightModeliv",
   "glLightf",
   "glLightfv",
   "glLighti",
   "glLightiv",
   "glLineStipple",
   "glLineWidth",
   "glListBase",
   "glLoadIdentity",
   "glLoadMatrixd",
   "glLoadMatrixf",
   "glLoadName",
   "glLogicOp",
   "glMap1d",
   "glMap1f",
   "glMap2d",
   "glMap2f",
   "glMapGrid1d",
   "glMapGrid1f",
   "glMapGrid2d",
   "glMapGrid2f",
   "glMaterialf",
   "glMaterialfv",
   "glMateriali",
   "glMaterialiv",
   "glMatrixMode",
   "glMultMatrixd",
   "glMultMatrixf",
   "glNewList",
   "glNormal3b",
   "glNormal3bv",
   "glNormal3d",
   "glNormal3dv",
   "glNormal3f",
   "glNormal3fv",
   "glNormal3i",
   "glNormal3iv",
   "glNormal3s",
   "glNormal3sv",
   "glNormalPointer",
   "glOrtho",
   "glPassThrough",
   "glPixelMapfv",
   "glPixelMapuiv",
   "glPixelMapusv",
   "glPixelStoref",
   "glPixelStorei",
   "glPixelTransferf",
   "glPixelTransferi",
   "glPixelZoom",
   "glPointSize",
   "glPolygonMode",
   "glPolygonOffset",
   "glPolygonStipple",
   "glPopAttrib",
   "glPopClientAttrib",
   "glPopMatrix",
   "glPopName",
   "glPrioritizeTextures",
   "glPushAttrib",
   "glPushClientAttrib",
   "glPushMatrix",
   "glPushName",
   "glRasterPos2d",
   "glRasterPos2dv",
   "glRasterPos2f",
   "glRasterPos2fv",
   "glRasterPos2i",
   "glRasterPos2iv",
   "glRasterPos2s",
   "glRasterPos2sv",
   "glRasterPos3d",
   "glRasterPos3dv",
   "glRasterPos3f",
   "glRasterPos3fv",
   "glRasterPos3i",
   "glRasterPos3iv",
   "glRasterPos3s",
   "glRasterPos3sv",
   "glRasterPos4d",
   "glRasterPos4dv",
   "glRasterPos4f",
   "glRasterPos4fv",
   "glRasterPos4i",
   "glRasterPos4iv",
   "glRasterPos4s",
   "glRasterPos4sv",
   "glReadBuffer",
   "glReadPixels",
   "glRectd",
   "glRectdv",
   "glRectf",
   "glRectfv",
   "glRecti",
   "glRectiv",
   "glRects",
   "glRectsv",
   "glRenderMode",
   "glRotated",
   "glRotatef",
   "glScaled",
   "glScalef",
   "glScissor",
   "glSelectBuffer",
   "glShadeModel",
   "glStencilFunc",
   "glStencilMask",
   "glStencilOp",
   "glTexCoord1d",
   "glTexCoord1dv",
   "glTexCoord1f",
   "glTexCoord1fv",
   "glTexCoord1i",
   "glTexCoord1iv",
   "glTexCoord1s",
   "glTexCoord1sv",
   "glTexCoord2d",
   "glTexCoord2dv",
   "glTexCoord2f",
   "glTexCoord2fv",
   "glTexCoord2i",
   "glTexCoord2iv",
   "glTexCoord2s",
   "glTexCoord2sv",
   "glTexCoord3d",
   "glTexCoord3dv",
   "glTexCoord3f",
   "glTexCoord3fv",
   "glTexCoord3i",
   "glTexCoord3iv",
   "glTexCoord3s",
   "glTexCoord3sv",
   "glTexCoord4d",
   "glTexCoord4dv",
   "glTexCoord4f",
   "glTexCoord4fv",
   "glTexCoord4i",
   "glTexCoord4iv",
   "glTexCoord4s",
   "glTexCoord4sv",
   "glTexCoordPointer",
   "glTexEnvf",
   "glTexEnvfv",
   "glTexEnvi",
   "glTexEnviv",
   "glTexGend",
   "glTexGendv",
   "glTexGenf",
   "glTexGenfv",
   "glTexGeni",
   "glTexGeniv",
   "glTexImage1D",
   "glTexImage2D",
   "glTexParameterf",
   "glTexParameterfv",
   "glTexParameteri",
   "glTexParameteriv",
   "glTexSubImage1D",
   "glTexSubImage2D",
   "glTranslated",
   "glTranslatef",
   "glVertex2d",
   "glVertex2dv",
   "glVertex2f",
   "glVertex2fv",
   "glVertex2i",
   "glVertex2iv",
   "glVertex2s",
   "glVertex2sv",
   "glVertex3d",
   "glVertex3dv",
   "glVertex3f",
   "glVertex3fv",
   "glVertex3i",
   "glVertex3iv",
   "glVertex3s",
   "glVertex3sv",
   "glVertex4d",
   "glVertex4dv",
   "glVertex4f",
   "glVertex4fv",
   "glVertex4i",
   "glVertex4iv",
   "glVertex4s",
   "glVertex4sv",
   "glVertexPointer",
   "glViewport",
   "wglChoosePixelFormat",
   "wglCopyContext",
   "wglCreateContext",
   "wglCreateLayerContext",
   "wglDeleteContext",
   "wglDescribeLayerPlane",
   "wglDescribePixelFormat",
   "wglGetCurrentContext",
   "wglGetCurrentDC",
   "wglGetDefaultProcAddress",
   "wglGetLayerPaletteEntries",
   "wglGetPixelFormat",
   "wglGetProcAddress",
   "wglMakeCurrent",
   "wglRealizeLayerPalette",
   "wglSetLayerPaletteEntries",
   "wglSetPixelFormat",
   "wglShareLists",
   "wglSwapBuffers",
   "wglSwapLayerBuffers",
   "wglUseFontBitmapsA",
   "wglUseFontBitmapsW",
   "wglUseFontOutlinesA",
   "wglUseFontOutlinesW"
};

//****************************************************************************
//*                                                                          *
//*  Return rectangle containing client-area boundaries in screenspace       *
//*                                                                          *
//****************************************************************************

static RECT *client_screen_rect(WINDOW *W)
{
   static RECT rect;
   POINT        ul,lr;

   GetClientRect(W->hWnd, &rect);

   ul.x = rect.left; 
   ul.y = rect.top; 
   lr.x = rect.right; 
   lr.y = rect.bottom; 

   if (GetMenu(W->hWnd) != NULL)
      {
      ul.y -= GetSystemMetrics(SM_CYMENU);
      }

   ClientToScreen(W->hWnd, &ul); 
   ClientToScreen(W->hWnd, &lr); 

   SetRect(&rect, ul.x, ul.y, 
                  lr.x-1, lr.y-1); 

   return &rect;
}

//****************************************************************************
//*                                                                          *
//*  Configure window for fullscreen mode display                            *
//*                                                                          *
//****************************************************************************

static void set_fullscreen_mode(WINDOW *W)
{
   HWND hWnd = W->hWnd;

   //
   // Set flag to indicate fullscreen transition in progress
   //
   // This will suppress our handling of WM_WINDOWPOSCHANGED, so that
   // transitioning from windowed to fullscreen will not reset the client 
   // area to the entire desktop.
   //

   ++W->setting_fullscreen_mode;

   //
   // Disable caption menu
   //

   SetWindowLong(hWnd, 
                 GWL_STYLE, 
                 GetWindowLong(hWnd, GWL_STYLE) | WS_POPUP);

   SetWindowLong(hWnd, 
                 GWL_STYLE, 
                 GetWindowLong(hWnd, GWL_STYLE) & ~(WS_OVERLAPPED  | 
                                                    WS_CAPTION     | 
                                                    WS_SYSMENU     | 
                                                    WS_MINIMIZEBOX | 
                                                    WS_MAXIMIZEBOX | 
                                                    WS_THICKFRAME));

   //
   // Get desktop size
   //

   S32 desktop_w = GetSystemMetrics(SM_CXSCREEN);
   S32 desktop_h = GetSystemMetrics(SM_CYSCREEN);

   //
   // Set window boundaries to cover entire desktop, and show it
   //

   SetWindowPos(hWnd, 
                HWND_TOP, 
                0,
                0,
                desktop_w,
                desktop_h,
                SWP_NOCOPYBITS | SWP_NOZORDER);

   ShowWindow(hWnd, SW_SHOWNORMAL);

   --W->setting_fullscreen_mode;
}

//****************************************************************************
//*                                                                          *
//*  Configure window for windowed-mode display                              *
//*                                                                          *
//****************************************************************************

static void set_windowed_mode(WINDOW          *W, //)
                              S32 display_width,
                              S32 display_height)
{
   S32  retry;
   RECT client_rect;
   RECT window_rect;

   HWND hWnd = W->hWnd;

   //
   // Get desktop size
   //

   S32 desktop_w = GetSystemMetrics(SM_CXSCREEN);
   S32 desktop_h = GetSystemMetrics(SM_CYSCREEN);

   //
   // Enable caption menu and user preferences
   //

   SetWindowLong(hWnd, 
                 GWL_STYLE, 
                 GetWindowLong(hWnd, GWL_STYLE) & ~WS_POPUP);

   SetWindowLong(hWnd, 
                 GWL_STYLE, 
                 GetWindowLong(hWnd, GWL_STYLE) | (WS_OVERLAPPED  | 
                                                   WS_CAPTION     | 
                                                   WS_SYSMENU     | 
                                                   WS_MINIMIZEBOX));

   if (0) //SAL_get_preference(SAL_ALLOW_WINDOW_RESIZE))
      {
      SetWindowLong(hWnd, 
                    GWL_STYLE, 
                    GetWindowLong(hWnd, GWL_STYLE) | WS_THICKFRAME |
                                                     WS_MAXIMIZEBOX);
      }

   if (0) //SAL_get_preference(SAL_ALWAYS_ON_TOP))
      {
      SetWindowLong(hWnd, 
                    GWL_EXSTYLE, 
                    GetWindowLong(hWnd, GWL_EXSTYLE) | WS_EX_TOPMOST);
      }
   else
      {
      SetWindowLong(hWnd, 
                    GWL_EXSTYLE, 
                    GetWindowLong(hWnd, GWL_EXSTYLE) & ~WS_EX_TOPMOST);
      }

   //
   // If area not already established, center window's client area on
   // desktop, and size it to correspond to the display size for optimum 
   // performance (no stretching needed)
   //

   if (W->area.w == -1)
      {
      W->area.w = display_width;
      W->area.h = display_height;

      W->area.x = ((desktop_w - W->area.w ) / 2);
      W->area.y = ((desktop_h - W->area.h) / 2);
      }

   //
   // Calculate adjusted position of window
   //
   // Do not allow overall window size to exceed desktop size; keep 
   // dividing height and width by 2 until entire window fits
   //
   // If window is offscreen (or almost entirely offscreen), center it
   //

   do
      {
      retry = 0;

      window_rect.left   = W->area.x;
      window_rect.right  = W->area.x + W->area.w - 1;
      window_rect.top    = W->area.y;
      window_rect.bottom = W->area.y + W->area.h - 1;

      //
      // Don't adjust window rect to compensate for a menu -- SAL is
      // designed to treat menu bars as if they were part of the
      // client area, for refresh and window-sizing purposes
      //

      AdjustWindowRectEx(&window_rect,
                          GetWindowLong(hWnd, GWL_STYLE),
                          0, //(GetMenu(hWnd) != NULL),
                          GetWindowLong(hWnd, GWL_EXSTYLE));

      if ((window_rect.right - window_rect.left + 1) > desktop_w)
         {
         W->area.w >>= 1;
         W->area.x = ((desktop_w - W->area.w ) / 2);
         retry = 1;
         }

      if ((window_rect.bottom - window_rect.top + 1) > desktop_h)
         {
         W->area.h >>= 1;
         W->area.y = ((desktop_h - W->area.h) / 2);
         retry = 1;
         }

      if ((window_rect.left   >= (desktop_w-16)) ||
          (window_rect.top    >= (desktop_h-16)) ||
          (window_rect.right  <= 16)          ||
          (window_rect.bottom <= 16))
         {
         W->area.x = ((desktop_w - W->area.w ) / 2);
         W->area.y = ((desktop_h - W->area.h) / 2);
         retry = 1;
         }
      }
   while (retry);

   //
   // Save window and client areas for restoration if maximize button pressed
   //

   W->original_window_rect = window_rect;
   W->original_area        = W->area;

   //
   // Set window size and position
   //

   SetWindowPos(hWnd, 
                HWND_TOP, 
                window_rect.left,
                window_rect.top,
                window_rect.right  - window_rect.left + 1,
                window_rect.bottom - window_rect.top  + 1,
                SWP_NOCOPYBITS | SWP_NOZORDER);

   GetClientRect(hWnd, 
                &client_rect);

   debug_printf("W32GL: Window at (%d,%d), client size = (%d,%d)\n",
      window_rect.left,
      window_rect.top,
      client_rect.right,
      client_rect.bottom);

   //
   // Make window visible
   //

   ShowWindow(hWnd, SW_SHOWNORMAL);
}

//****************************************************************************
//*                                                                          *
//*  Window message receiver procedure                                       *
//*                                                                          *
//****************************************************************************

long FAR PASCAL WndProc(HWND   hWindow, UINT   message,   //)
                        WPARAM wParam,  LPARAM lParam)
{
   //
   // Get our window descriptor
   //
   // The search operator will fail during the first CreateWindow() call,
   // before WINDOW has been allocated
   //

   HWINDOW wnd = (HWINDOW) windows.search(hWindow);

   WINDOW *W = NULL;
   S32     is_primary_window = 1;

   if (wnd != -1)
      {
      W = &windows.list[wnd];
      is_primary_window = W->flags & UCW_PRIMARY;
      }

   //
   // Handle message
   // 

   switch (message)
      {
      case WM_ACTIVATEAPP:

         if ((!is_primary_window) || (W == NULL))
            {
            break;
            }

         if (wParam)
            {
            if (!app_active)
               {
               app_active = TRUE;
               debug_printf("W32GL: Activated!\n");

               if (W->window_mode == UWM_FULLSCREEN)
                  {
                  SetWindowLong(hWindow, 
                                GWL_STYLE, 
                                GetWindowLong(hWindow, 
                                              GWL_STYLE) & ~WS_SYSMENU);
                  }
               }
            }
         else
            {
            if (app_active)
               {
               app_active = FALSE;
               debug_printf("W32GL: Deactivated!\n");

               if (W->window_mode == UWM_FULLSCREEN)
                  {
                  SetWindowLong(hWindow, 
                                GWL_STYLE, 
                                GetWindowLong(hWindow, 
                                              GWL_STYLE) | WS_SYSMENU);
                  }
               }
            }
         break;

      case WM_SYSCOMMAND:

         switch (wParam & 0xffff0)
            {
            case SC_SCREENSAVE:
            case SC_MONITORPOWER:

               //
               // Inhibit all screen-saver activity
               //

               return 0;
            }

         break;

      case WM_SYSKEYUP:

         if (wParam == VK_MENU)
            {
            if (1) //SAL_get_preference(SAL_PREVENT_ALT_MENU_POPUP))
               {
               //
               // If key pressed was ALT by itself, set wParam to 0 (invalid
               // key code) to prevent system menu from popping up and 
               // interfering with app
               //

               wParam = 0;
               }
            }
         break;

      case WM_DESTROY:

         if (!is_primary_window)
            {
            break;
            }

         debug_printf("W32GL: Destroy message received\n");
         PostQuitMessage(0);
         break;
      }

   if ((W != NULL) && (W->callback != NULL))
      {
      return (W->callback)(hWindow, message, wParam, lParam);
      }
   else
      {
      return DefWindowProc(hWindow, message, wParam, lParam);
      }
}

/*********************************************************************/
//
// Constructor/destructor
//
/*********************************************************************/

USERIO::USERIO()
{
}

USERIO::~USERIO()
{
}

/*********************************************************************/
//
// startup()
//
/*********************************************************************/

S32     USERIO::startup                      (HINSTANCE  instance, //)
                                              C8        *app_name,
                                              S32        allow_multiple_instances,
                                              USEREXITCB exit_handler)
{
   //
   // Generate startup message
   //

   debug_printf("Starting W32GL (%s): %s, HINSTANCE %X\n",
      __DATE__,
      app_name,
      instance);

   //
   // If desired, reject any attempt to launch more than one instance of this 
   // application
   //

   if (!allow_multiple_instances)
      {
      //
      // Check for redundant startup by creating a semaphore and checking to
      // see if it already existed
      //
      // Semaphore will be destroyed automatically at shutdown time
      //

      HANDLE hSem = CreateSemaphore(NULL,     
                                    0,        
                                    1,        
                                    app_name);

      if ((hSem != NULL) && (GetLastError() == ERROR_ALREADY_EXISTS))
         {
         CloseHandle(hSem);

         debug_printf("W32GL: Multiple instances not permitted, creation failed\n");
         return -1;
         }
      }

   //             
   // Set global instance/app name variables
   // 

   hAppInstance = instance;

   strncpy(szAppName,
           app_name,
           sizeof(szAppName));

   //
   // Set (mandatory) exit callback function
   //

   if (exit_handler == NULL)
      {
      return 0;
      }

   exit_callback = exit_handler;


   //
   // Init globals 
   //

   app_active =  FALSE;  // App not activated yet

   //
   // Indicate successful startup
   //

   return 1;
}

/*********************************************************************/
//
// shutdown()
//
/*********************************************************************/

void    USERIO::shutdown                     (void)
{
   //
   // Manually destroy all windows
   //

   for (S32 i=0; i < windows.list_size; i++)
      {
      WINDOW *W = &windows.list[i];

      if (W->index == -1)
         {
         continue;
         }

      close_window(i);
      }

   //
   // Free all allocated windows
   // 

   windows.free();
}

/*********************************************************************/
//
// is_app_active()
//
/*********************************************************************/

S32          USERIO::is_app_active           (void)
{
   return app_active;
}

/*********************************************************************/
//
// show_system_mouse()
//
/*********************************************************************/

void         USERIO::show_system_mouse       (void)
{
   ShowCursor(1);
}

/*********************************************************************/
//
// hide_system_mouse()
//
/*********************************************************************/

void         USERIO::hide_system_mouse       (void)
{
   ShowCursor(0);
   debug_printf("Hidden\n");
}

/*********************************************************************/
//
// open_window()  
//
/*********************************************************************/

HWINDOW USERIO::open_window          (S32         display_width, //)
                                      S32         display_height,
                                      S32         display_bpp,
                                      USERWINMODE window_mode,
                                      C8         *GL_provider,
                                      GLVTABLE  **GLV,
                                      U32         flags)
{
   //
   // Set up and register application window class
   //

   WNDCLASS wc;

   wc.style         = CS_HREDRAW | CS_VREDRAW;
   wc.lpfnWndProc   = WndProc;
   wc.cbClsExtra    = 0;
   wc.cbWndExtra    = 0;
   wc.hInstance     = hAppInstance;
   wc.hIcon         = LoadIcon(hAppInstance, IDI_APPLICATION);
   wc.hCursor       = LoadCursor(NULL, IDC_ARROW);
   wc.hbrBackground = (HBRUSH) GetStockObject(BLACK_BRUSH);
   wc.lpszMenuName  = NULL;
   wc.lpszClassName = szAppName;

   RegisterClass(&wc);

   //
   // Create application's main window
   //

   HWND hWnd = CreateWindowEx(
       0,
       szAppName,
       szAppName,
       0,
       CW_USEDEFAULT,
       CW_USEDEFAULT,
       CW_USEDEFAULT,
       CW_USEDEFAULT,
       NULL,
       NULL,
       hAppInstance,
       NULL);

   if (!hWnd)
      {
      debug_printf("W32GL: Invalid window handle!\n");
      return -1;
      }

   UpdateWindow(hWnd);
   SetFocus(hWnd);

   //
   // Allocate and initialize window descriptor
   //

   HWINDOW wnd = windows.allocate(hWnd);

   WINDOW *W = &windows.list[wnd];

   W->setting_fullscreen_mode = 0;
   W->window_mode             = window_mode;
   W->display_width           = display_width;
   W->display_height          = display_height;
   W->display_bpp             = display_bpp;

   W->flags                   = flags;
   W->callback                = NULL;
   W->area.w                  = -1;

   W->quit_message_received   = FALSE;
   W->manually_destroying     = FALSE;

   //
   // Clear OpenGL state
   //

   memset(&W->GL,
           0,
           sizeof(W->GL));

   W->GL_name[0]              = 0;
   W->GL_RC                   = NULL;
   W->GL.hDC                  = NULL;
   W->GL_fullscreen           = 0;

   //
   // Load requested OpenGL provider
   //

   W->GL.hProvider = LoadLibrary(GL_provider);

   if (W->GL.hProvider == NULL)
      {
      debug_printf("W32GL: OpenGL provider not loaded, GetLastError() = 0x%X\n",GetLastError());
      return -1;
      }

   strcpy(W->GL_name, 
          GL_provider);

   //
   // Map Win32 OpenGL calls to our jump table
   //

   for (S32 i=0; i < ARY_CNT(GLN); i++)
      {
      FARPROC proc = GetProcAddress(W->GL.hProvider, GLN[i]);

      assert(proc != NULL);

      ((U32 *) &W->GL)[i] = (U32) proc;
      }

   //
   // Modify window parameters according to desired mode
   //

   if (window_mode == UWM_WINDOW)
      {
      set_windowed_mode(W, 
                        display_width,
                        display_height);

      W->GL_fullscreen = 0;
      }
   else
      {
      //
      // Disable lame-ass 3Dfx splash screen
      // 

      _putenv("FX_GLIDE_NO_SPLASH=1");

      //
      // Locate matching fullscreen display mode
      //

      DEVMODE dmode;
      S32     modenum = 0;
      BOOL    mode_found = TRUE;

      while (mode_found)
         {
         mode_found = EnumDisplaySettings(NULL, modenum, &dmode);

         if (mode_found                                  && 
            (dmode.dmBitsPerPel == (U32) display_bpp)    &&
            (dmode.dmPelsWidth  == (U32) display_width) &&
            (dmode.dmPelsHeight == (U32) display_height)) 
            {
            break;
            }

         modenum++;
         } 

      if (!mode_found)
         {
         debug_printf("W32GL: EnumDisplaySettings() failed to locate requested mode\n");
         return -1;
         }

      dmode.dmFields = DM_PELSWIDTH | DM_PELSHEIGHT | DM_BITSPERPEL;

      //
      // Try to set it
      //

      S32 changeResult = ChangeDisplaySettings(&dmode, 
                                                CDS_TEST);

      switch (changeResult)
         {
         case DISP_CHANGE_SUCCESSFUL:

            ChangeDisplaySettings (&dmode, CDS_FULLSCREEN);
            break;

         case DISP_CHANGE_FAILED:
         case DISP_CHANGE_BADMODE:

            debug_printf ("W32GL: Failed to change mode to desired settings, result = 0x%X\n",changeResult);
            return -1;

         case DISP_CHANGE_RESTART:

            debug_printf ("W32GL: Mode change failed (would require restart)\n");
            return -1;
         }

      //
      // We must change the window settings for fullscreen mode AFTER changing
      // the display size...
      //

      set_fullscreen_mode(W);

      W->GL_fullscreen = 1;
      }

   //
   // Get DC for window (required for some of the app's GL calls)
   //

   W->GL.hDC = GetDC(W->hWnd);

   //
   // Find best available match for requested pixel format
   //

   static PIXELFORMATDESCRIPTOR pfd = 
      {
      sizeof(PIXELFORMATDESCRIPTOR), // size of this pfd
      1,                             // version number
      PFD_DRAW_TO_WINDOW             // support window
    | PFD_SUPPORT_OPENGL             // support OpenGL
    | PFD_DOUBLEBUFFER,              // double buffered
      PFD_TYPE_RGBA,                 // RGBA type
 (U8) display_bpp,                   // color depth
      0, 0, 0, 0, 0, 0,              // color bits ignored
      0,                             // no alpha buffer
      0,                             // shift bit ignored
      0,                             // no accumulation buffer
      0, 0, 0, 0,                    // accum bits ignored
      16,                            // 16-bit z-buffer      
      0,                             // no stencil buffer
      0,                             // no auxiliary buffer
      PFD_MAIN_PLANE,                // main layer
      0,                             // reserved
      0, 0, 0                        // layer masks ignored
      };

   S32 pixelFormat = ChoosePixelFormat(W->GL.hDC, 
                                      &pfd);

   if (pixelFormat == 0)
      {
      debug_printf("W32GL: ChoosePixelFormat() failed, GetLastError() = 0x%X\n",GetLastError());
      return -1;
      }

   debug_printf("W32GL: Setting pixel format #%d\n",pixelFormat);

   //
   // Associate valid pixel format with main window's device context
   //
   // (NT2K/Riva TNT2 fails with INVALID_PIXEL_FORMAT if we use W->GL.wglSetPixelFormat here)
   //

   if (!SetPixelFormat(W->GL.hDC, 
                       pixelFormat, 
                      &pfd))
      {
      debug_printf("W32GL: SetPixelFormat() failed, GetLastError() = 0x%X\n",GetLastError());
      return -1;
      }

   //
   // Create OpenGL rendering context and associate it with the current thread
   //

   W->GL_RC = W->GL.wglCreateContext(W->GL.hDC);

   if (W->GL_RC == NULL)
      {
      debug_printf("W32GL: wglCreateContext() failed, GetLastError() = 0x%X\n",GetLastError());
      return -1;
      }

   if (!W->GL.wglMakeCurrent(W->GL.hDC, 
                             W->GL_RC))
      {
      W->GL.wglDeleteContext(W->GL_RC);
      W->GL_RC = NULL;

      debug_printf("W32GL: wglMakeCurrent() failed, GetLastError() = 0x%X\n",GetLastError());
      return -1;
      }

   //
   // Set global flag to indicate OpenGL is active, and return success
   // 

   debug_printf("W32GL: Initialized OpenGL provider %s, GL_VENDOR = %s\n",
      W->GL_name,
      W->GL.glGetString(GL_VENDOR));

   //
   // Return window handle to application
   //

   if (flags & UCW_PRIMARY)
      {
      app_active = TRUE;
      }

   *GLV = &W->GL;

   return wnd;
}

/*********************************************************************/
//
// close_window()  
//
/*********************************************************************/

void    USERIO::close_window               (HWINDOW window)
{
   if (window == -1)
      {
      return;
      }

   WINDOW *W = &windows.list[window];

   //
   // Destroy GL context and restore display mode
   //

   debug_printf("W32GL: Closing window %d\n",window);

   W->GL.wglMakeCurrent(0,0);

   if (W->GL_fullscreen)
      {
      ChangeDisplaySettings(NULL, 0);
      W->GL_fullscreen = 0;
      }

   if (W->GL_RC != NULL)
      {
      W->GL.wglDeleteContext(W->GL_RC);
      W->GL_RC = NULL;
      }

   if (W->GL.hDC != NULL)
      {
      ReleaseDC(W->hWnd, W->GL.hDC);
      W->GL.hDC = NULL;
      }

   //
   // Destroy the window, if it hasn't already been manually closed by the 
   // user (e.g., by hitting ALT-F4)
   //

   HWND hWnd = W->hWnd;

   if (!W->quit_message_received)
      {
      W->manually_destroying = 1;

      DestroyWindow(hWnd);

      while (!W->quit_message_received)
         {
         serve_window(window);
         Sleep(10);
         }

      W->manually_destroying = 0;
      W->quit_message_received = 0;
      }

   //
   // Unload OpenGL provider library
   //
   // We have to do this AFTER destroying the window, since some buggy GL
   // drivers don't release their window hook when their rendering context is 
   // deleted
   // 

   if (W->GL.hProvider != NULL)
      {
      debug_printf("W32GL: Unloading OpenGL provider %s\n",W->GL_name);
 
      FreeLibrary(W->GL.hProvider);
 
      memset(&W->GL,
             0,
             sizeof(W->GL));
      }

   //
   // Free our window descriptor
   //

   windows.unlink(window);
}

/*********************************************************************/
//
// get_system_window_handle()
//
/*********************************************************************/

HWND    USERIO::get_system_window_handle     (HWINDOW window)
{
   if (window == -1)
      {
      return NULL;
      }

   WINDOW *W = &windows.list[window];

   HWND hWnd = W->hWnd;

   return hWnd;
}

/*********************************************************************/
//
// register_system_window_proc()
//
/*********************************************************************/

WNDPROC USERIO::register_system_window_proc  (HWINDOW window, //)
                                              WNDPROC fn)
{
   if (window == -1)
      {
      return NULL;
      }

   WINDOW *W = &windows.list[window];

   WNDPROC old;

   old = W->callback;

   W->callback = fn;

   return old;
}

/*********************************************************************/
//
// serve_window()
//
/*********************************************************************/

void        USERIO::serve_window   (HWINDOW window)
{
   WINDOW *W = &windows.list[window];

   if (window == -1)
      {
      return;
      }

   HWND hWnd = W->hWnd;

   MSG msg;

   //
   // Serve message queue
   //

   while (PeekMessage(&msg, NULL, 0, 0, PM_NOREMOVE))
      {
      if (!GetMessage(&msg, NULL, 0, 0 ))
         {
         debug_printf("Quit message received\n");
         W->quit_message_received = TRUE;

         if (!W->manually_destroying)
            {
            //
            // If we are not already in close_window(), call the user 
            // exit routine to clean up the app and destroy all windows
            //

            (exit_callback)();
            }

         return;
         }

      TranslateMessage(&msg); 
      DispatchMessage(&msg);
      }
}

/*********************************************************************/
//
// get_system_GL_provider()
//
/*********************************************************************/

GLVTABLE *USERIO::get_system_GL_provider  (HWINDOW window)
{
   WINDOW *W = &windows.list[window];

   if (window == -1)
      {
      return NULL;
      }

   return &W->GL;
}

/*********************************************************************/
//
// window_statistics()
//
/*********************************************************************/

void USERIO::window_statistics         (HWINDOW      window, //)
                                        S32         *display_width,
                                        S32         *display_height,
                                        S32         *display_bpp, 
                                        USERWINMODE *window_mode)
{
   WINDOW *W = &windows.list[window];

   if (window == -1)
      {
      return;
      }

   if (display_width  != NULL) *display_width  = W->display_width;
   if (display_height != NULL) *display_height = W->display_height;
   if (display_bpp    != NULL) *display_bpp    = W->display_bpp;
   if (window_mode    != NULL) *window_mode    = W->window_mode;
}

/*********************************************************************/
//
// client_area()
//
/*********************************************************************/

void USERIO::client_area     (HWINDOW window, //)
                              WINAREA *area)
{
   WINDOW *W = &windows.list[window];

   if (window == -1)
      {
      return;
      }

   RECT *rect;

   if (W->window_mode == UWM_WINDOW)
      {
      rect = client_screen_rect(W);

      area->x = rect->left;
      area->y = rect->top;
      area->w = rect->right  - rect->left + 1;
      area->h = rect->bottom - rect->top  + 1;
      }
   else
      {
      area->x = 0;
      area->y = 0;
      area->w = W->display_width;
      area->h = W->display_height;
      }
}
